==> japhar-0.04.tar.gz: japhar-0.04/lib/libruntime/arrays.c <==
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "ClazzFile.h"
#include "arrays.h"
#include "array-class.h"
#include "objects.h"
#include "resolve.h"
#include "log.h"
#include "interp.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define ARRAY_LOG "ArrayRef"

static void
array_finalize(GC_obj *obj)
{
  obj_thunk *array = (obj_thunk*)obj;
  void *elements = INSTANCE_DATA(array)->fields[ARRAY_BODY_INDEX].value.l;
#if 0
  SigPrimType element_type = INSTANCE_DATA(array)->fields[ARRAY_TYPE_INDEX].value.i;

  printf ("Finalizing array %p or type %s\n", obj, SIG_formatPrimitiveTypeToC(THREAD_getEnv(), element_type));
#endif

  /* we need to call the finalizer (since the java code may need the elements)
     before we free the body of the array... do we *really* need to do this,
     I mean, can user code put a finalizer method into an array class? */

  /* object_finalize(obj); */

  if (elements)
    free(elements);

  free(array - 1);
}

static GC_obj gc_template = {
  NULL, NULL,

  GC_FLAG_ARRAY,

  sizeof(obj_thunk),

  array_finalize
};

obj_thunk *
new_array(JNIEnv *env, u4 n_elem, ClazzFile *array_type)
{
  obj_thunk *new_arr = new_object(env, array_type);
  int el_size = 0;
  char *array_class_name = getClassName(env, array_type);
  JavaStackItem *field;
  int el_type;

  memcpy(&new_arr->_gc_obj, &gc_template, sizeof(GC_obj));

  switch (array_class_name[1])
    {
    case '[':
      el_type = SIG_JOBJECT;
      el_size = sizeof(jobject);
      break;
    case 'L':
      el_type = SIG_JOBJECT;
      el_size = sizeof(jobject);
      break;
    case 'I':
      el_type = SIG_JINT;
      el_size = sizeof(jobject);
      break;
    case 'F':
      el_type = SIG_JFLOAT;
      el_size = sizeof(jfloat);
      break;
    case 'J':
      el_type = SIG_JLONG;
      el_size = sizeof(jlong);
      break;
    case 'D':
      el_type = SIG_JDOUBLE;
      el_size = sizeof(jdouble);
      break;
    case 'C':
      el_type = SIG_JCHAR;
      el_size = sizeof(jchar);
      break;
    case 'S':
      el_type = SIG_JSHORT;
      el_size = sizeof(jshort);
      break;
    case 'B':
      el_type = SIG_JBYTE;
      el_size = sizeof(jbyte);
      break;
    case 'Z':
      el_type = SIG_JBOOLEAN;
      el_size = sizeof(jboolean);
      break;
    default:
      /* XXX punt */
      el_type = SIG_JINT;
      el_size = sizeof(jint);
      break;
    }

  /* do the length of the array */
  field = &(INSTANCE_DATA(new_arr)->fields[ARRAY_LENGTH_INDEX]);
  field->value.i = n_elem;

  /* do the array's type */
  field = &(INSTANCE_DATA(new_arr)->fields[ARRAY_TYPE_INDEX]);
  field->value.i = el_type;

  /* do the elsize */
  field = &(INSTANCE_DATA(new_arr)->fields[ARRAY_ELSIZE_INDEX]);
  field->value.i = el_size;

  /* do the body of the array */
  field = &(INSTANCE_DATA(new_arr)->fields[ARRAY_BODY_INDEX]);
  if (n_elem == 0)
    field->value.l = (void*)calloc(1, el_size); /* just so we have *something* */
  else
    field->value.l = (void*)calloc(n_elem, el_size);

  /* we need to change the finalize method so that we can free our elements.  we promise
     to call the object_finalize method. */
  new_arr->_gc_obj.finalize = array_finalize;

  JAVARLOG5 (ARRAY_LOG, 1, "new array --> %p (%ld) {body = %p(%ld), length = %d}\n",
	     new_arr, (long)new_arr,
	     INSTANCE_DATA(new_arr)->fields[ARRAY_BODY_INDEX].value.l,
	     INSTANCE_DATA(new_arr)->fields[ARRAY_BODY_INDEX].value.i,
	     INSTANCE_DATA(new_arr)->fields[ARRAY_LENGTH_INDEX].value.i);

  return new_arr;
}


obj_thunk *
multi_new_array (JNIEnv *env, jint *n_elems, int dimensions, ClazzFile *array_type)
{
  int i;
  obj_thunk *new_arr = new_array (env, n_elems[ 0 ], array_type);
  JavaStackItem body_item;

  get_instance_field(new_arr,
		     STATIC_DATA(new_arr)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  if (dimensions > 1)
    {
      ClazzFile *subarray_type = createFakeArrayRemoveDimension(env, array_type);
      
      for (i=0; i<n_elems[ 0 ]; i++)
	((jobject*)body_item.value.l)[ i ] =
	  (jobject) multi_new_array (env, n_elems + 1, dimensions - 1, subarray_type);
    }

  return new_arr;
}

==> japhar-0.05.tar.gz: japhar-0.05/lib/libruntime/arrays.c <==
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "ClazzFile.h"
#include "arrays.h"
#include "array-class.h"
#include "objects.h"
#include "resolve.h"
#include "log.h"
#include "interp.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define ARRAY_LOG "ArrayRef"

static void
array_finalize(GC_obj *obj)
{
#if 0
  japhar_object* array = (japhar_object*)obj;
  jvalue v;
  void *elements;

  SigPrimType element_type = INSTANCE_DATA(array)->fields[ARRAY_TYPE_INDEX].value.i;

  printf ("Finalizing array %p or type %s\n", obj, SIG_formatPrimitiveTypeToC(THREAD_getEnv(), element_type));

  get_instance_field(array->_static + 1,
		     array->_static,
		     INSTANCE_DATA(array)->fields[ARRAY_BODY_INDEX].value.l;

  /* we need to call the finalizer (since the java code may need the elements)
     before we free the body of the array... do we *really* need to do this,
     I mean, can user code put a finalizer method into an array class? */

  /* object_finalize(obj); */

  if (elements)
    free(elements);

  free(array - 1);
#endif
}

static GC_obj gc_template = {
  NULL, NULL,

  GC_FLAG_ARRAY,

  100, /* XXXXX */

  array_finalize
};

japhar_obj 
new_array(JNIEnv *env, u4 n_elem, ClazzFile *array_type)
{
  japhar_obj new_arr;
  japhar_object *arr_object;
  int el_size = 0;
  char *array_class_name = getClassName(env, array_type);
  FieldStruct *field;
  jvalue value;
  int el_type;

  new_arr = new_object(env, array_type);

  obj_to_object(new_arr, arr_object);

  memcpy(&arr_object->_gc_obj, &gc_template, sizeof(GC_obj));

  switch (array_class_name[1])
    {
    case '[':
      el_type = SIG_JOBJECT;
      el_size = sizeof(jobject);
      break;
    case 'L':
      el_type = SIG_JOBJECT;
      el_size = sizeof(jobject);
      break;
    case 'I':
      el_type = SIG_JINT;
      el_size = sizeof(jobject);
      break;
    case 'F':
      el_type = SIG_JFLOAT;
      el_size = sizeof(jfloat);
      break;
    case 'J':
      el_type = SIG_JLONG;
      el_size = sizeof(jlong);
      break;
    case 'D':
      el_type = SIG_JDOUBLE;
      el_size = sizeof(jdouble);
      break;
    case 'C':
      el_type = SIG_JCHAR;
      el_size = sizeof(jchar);
      break;
    case 'S':
      el_type = SIG_JSHORT;
      el_size = sizeof(jshort);
      break;
    case 'B':
      el_type = SIG_JBYTE;
      el_size = sizeof(jbyte);
      break;
    case 'Z':
      el_type = SIG_JBOOLEAN;
      el_size = sizeof(jboolean);
      break;
    default:
      /* XXX punt */
      el_type = SIG_JINT;
      el_size = sizeof(jint);
      break;
    }

  /* do the length of the array */
  field = (*new_arr)->fields [ ARRAY_LENGTH_INDEX ];
  value.i = n_elem;
  set_instance_field(new_arr, field, value);

  /* do the array's type */
  field = (*new_arr)->fields [ ARRAY_TYPE_INDEX ];
  value.i = el_type;
  set_instance_field(new_arr, field, value);

  /* do the elsize */
  field = (*new_arr)->fields [ ARRAY_ELSIZE_INDEX ];
  value.i = el_size;
  set_instance_field(new_arr, field, value);

  /* do the body of the array */
  field = (*new_arr)->fields [ ARRAY_BODY_INDEX ];
  if (n_elem == 0)
    value.l = (void*)calloc(1, el_size); /* just so we have *something* */
  else
    value.l = (void*)calloc(n_elem, el_size);
  set_instance_field(new_arr, field, value);

  JAVARLOG5 (ARRAY_LOG, 1, "new array --> %p (%ld) {body = %p(%ld), length = %d}\n",
	     new_arr, (long)new_arr, value.l, value.j, n_elem);

  return new_arr;
}


japhar_obj 
multi_new_array (JNIEnv *env, jint *n_elems, int dimensions, ClazzFile *array_type)
{
  int i;
  japhar_obj new_arr = new_array (env, n_elems[ 0 ], array_type);
  jvalue body_item;

  get_instance_field(new_arr,
		     (*new_arr)->fields [ ARRAY_BODY_INDEX ],
		     &body_item);

  if (dimensions > 1)
    {
      ClazzFile *subarray_type = createFakeArrayRemoveDimension(env, array_type);
      
      for (i=0; i<n_elems[ 0 ]; i++)
	((jobject*)body_item.l)[ i ] =
	  (jobject) multi_new_array (env, n_elems + 1, dimensions - 1, subarray_type);
    }

  return new_arr;
}
